import { Observable, MonoTypeOperatorFunction, OperatorFunction, Subscription, TeardownLogic, Observer, BehaviorSubject } from 'rxjs';
import { Container, Token } from 'ditox';

/**
 * Action is an event emitter
 *
 * @param operator Optional transformation or handler for an event
 *
 * @field event$ - Observable for emitted events.
 *
 * @example
 * ```ts
 * // Create the action
 * const submitForm = createAction<{login: string, password: string}>();
 *
 * // Call the action
 * submitForm({login: 'foo', password: 'bar'});
 *
 * // Handle action's events
 * submitForm.even$.subscribe((formData) => {
 *   // Process the formData
 * });
 * ```
 */
type Action<Event> = {
    readonly event$: Observable<Event>;
    (event: Event): void;
} & ([Event] extends [undefined | void] ? {
    (event?: Event): void;
} : {
    (event: Event): void;
});
declare function createAction<Event = void>(operator?: MonoTypeOperatorFunction<Event>): Action<Event>;

/**
 * Provider for a value of a state.
 */
type Query<T> = Readonly<{
    /** Returns the value of a state */
    get: () => T;
    /** `Observable` for value changes.  */
    value$: Observable<T>;
}>;

type EffectResult<Event, Value> = Readonly<{
    event: Event;
    result: Value;
}>;
type EffectErrorOrigin = 'source' | 'handler';
type EffectError<Event, ErrorType> = Readonly<{
    origin: 'source';
    event?: undefined;
    error: any;
} | {
    origin: 'handler';
    event: Event;
    error: ErrorType;
}>;
type EffectNotification<Event, Result, ErrorType> = Readonly<({
    type: 'result';
} & EffectResult<Event, Result>) | ({
    type: 'error';
} & EffectError<Event, ErrorType>)>;
/**
 * Details about performing the effect.
 */
type EffectState<Event, Result = void, ErrorType = Error> = Readonly<{
    /** Provides a result of successful execution of the handler */
    result$: Observable<Result>;
    /** Provides a source event and a result of successful execution of the handler */
    done$: Observable<EffectResult<Event, Result>>;
    /** Provides an error emitter by a source (`event` is `undefined`)
     * or by the handler (`event` is not `undefined`) */
    error$: Observable<EffectError<Event, ErrorType>>;
    /** Provides a notification after execution of the handler for both success or error result  */
    final$: Observable<EffectNotification<Event, Result, ErrorType>>;
    /** Provides `true` if there is any execution of the handler in progress */
    pending: Query<boolean>;
    /** Provides a count of the handler in progress */
    pendingCount: Query<number>;
}>;

type AnyObject = Record<string, any>;
/**
 * Makes shallow comparison of two objects.
 */
declare const OBJECT_COMPARATOR: (objA: Record<string, unknown>, objB: Record<string, unknown>) => boolean;

/**
 * Effects and business logic controller.
 *
 * Implementation of the controller must provide `destroy()` method. It should
 * be used for closing subscriptions and disposing resources.
 *
 * @example
 * ```ts
 * type LoggerController = Controller<{
 *   log: (message: string) => void;
 * }>;
 * ```
 */
type Controller<ControllerProps extends AnyObject = AnyObject> = Readonly<ControllerProps & {
    /** Dispose the controller and clean its resources */
    destroy: () => void;
}>;

declare const GLOBAL_EFFECT_UNHANDLED_ERROR$: Observable<EffectError<unknown, unknown>>;
type EffectController<Event, Result, ErrorType = Error> = Controller<{
    state: EffectState<Event, Result, ErrorType>;
    start: () => void;
    next: (result: EffectResult<Event, Result>) => void;
    complete: () => void;
    error: (error: EffectError<Event, ErrorType>) => void;
}>;
declare function createEffectController<Event, Result, ErrorType = Error>(): EffectController<Event, Result, ErrorType>;

/**
 * Handler for an event. It can be asynchronous.
 *
 * @result a result, Promise or Observable
 */
type EffectHandler<Event, Result> = (event: Event) => Result | Promise<Result> | Observable<Result>;
type EffectEventProject<Event, Result> = (event: Event) => Observable<Result>;
type EffectPipeline<Event, Result> = (eventProject: EffectEventProject<Event, Result>) => OperatorFunction<Event, Result>;
type EffectOptions<Event, Result> = Readonly<{
    /**
     * Custom pipeline for processing effect's events.
     *
     * `mergeMap` pipeline is used by default.
     */
    pipeline?: EffectPipeline<Event, Result>;
}>;
/**
 * Effect encapsulates a handler for Action or Observable.
 *
 * It provides the state of execution results, which can be used to construct
 * a graph of business logic.
 *
 * Effect collects all internal subscriptions, and provides `destroy()` methods
 * unsubscribe from them and deactivate the effect.
 */
type Effect<Event, Result = void, ErrorType = Error> = Controller<EffectState<Event, Result, ErrorType> & {
    handle: (source: Action<Event> | Observable<Event> | Query<Event>) => Subscription;
}>;
/**
 * Creates `Effect` from the provided handler.
 *
 * @example
 * ```ts
 * const sumEffect = createEffect<{a: number, b: number}, number>((event) => {
 *   return a + b;
 * });
 * ```
 */
declare function createEffect<Event = void, Result = void, ErrorType = Error>(handler: EffectHandler<Event, Result>, options?: EffectOptions<Event, Result>): Effect<Event, Result, ErrorType>;

/**
 * A function to update a state.
 *
 * It is recommended to return a new state or the previous one.
 *
 * Actually, the function can change the state in place, but it is responsible
 * for a developer to provide `comparator` function to the store which handles
 * the changes.
 *
 * For making changes use a currying function to provide arguments:
 * ```ts
 * const addPizzaToCart = (name: string): StateMutation<Array<string>> =>
 *   (state) => ([...state, name]);
 * ```
 *
 * @param state a previous state
 * @returns a next state
 */
type StateMutation<State> = (state: State) => State;
/**
 * A record of factories which create state mutations.
 */
type StateUpdates<State> = Readonly<Record<string, (...args: any[]) => StateMutation<State>>>;
/**
 * Declare a record of factories for creating state mutations.
 */
declare function declareStateUpdates<State>(): <Updates extends StateUpdates<State> = StateUpdates<State>>(updates: Updates) => Updates;
/**
 * Declare a record of factories for creating state mutations.
 */
declare function declareStateUpdates<State, Updates extends StateUpdates<State> = StateUpdates<State>>(stateExample: State, updates: Updates): Updates;
/**
 * Returns a mutation which applies all provided mutations for a state.
 *
 * You can use this helper to apply multiple changes at the same time.
 */
declare function pipeStateMutations<State>(mutations: ReadonlyArray<StateMutation<State>>): StateMutation<State>;
type StoreQueryFn<State> = {
    (): Query<State>;
    <T>(selector: (state: State) => T): Query<T>;
};
/**
 * Read-only interface of a store.
 */
type StoreQuery<State> = Readonly<Query<State> & {
    /**
     * Creates a query from the store.
     *
     * If selector is not specified, then the store is returned as is.
     *
     * If selector is specified then returns a part of the state as `Query`.
     * The result query produces distinct values.
     *
     * @example
     * ```ts
     * const state: StateReader<{form: {login: 'foo'}}> = // ...
     * const query = state.query((state) => state.form.login);
     * ```
     */
    query: StoreQueryFn<State>;
}>;
/**
 * @internal
 * Updates the state by provided mutations
 * */
type StoreUpdateFunction<State> = (mutation: StateMutation<State> | ReadonlyArray<StateMutation<State> | undefined | null | false>) => void;
/** Function which changes a state of the store */
type StoreUpdate<Args extends unknown[]> = (...args: Args) => void;
/** Record of store update functions */
type StoreUpdates<State, Updates extends StateUpdates<State>> = Readonly<{
    [K in keyof Updates]: StoreUpdate<Parameters<Updates[K]>>;
}>;
/**
 * Store of a state
 */
type Store<State> = Controller<StoreQuery<State> & {
    id: number;
    name?: string;
    /** Sets a new state to the store */
    set: (state: State) => void;
    /** Updates the store by provided mutations */
    update: StoreUpdateFunction<State>;
    /** Immediately emits a current state to store's subscribers */
    notify: () => void;
}>;
/** Store of a state with updating functions */
type StoreWithUpdates<State, Updates extends StateUpdates<State>> = Readonly<Store<State> & {
    updates: StoreUpdates<State, Updates>;
}>;
type StoreOptions<State> = Readonly<{
    name?: string;
    /** A comparator for detecting changes between old and new states */
    comparator?: (prevState: State, nextState: State) => boolean;
    /** Callback is called when the store is destroyed */
    onDestroy?: () => void;
}>;
/** @internal */
type InternalStoreOptions<State> = Readonly<StoreOptions<State> & {
    internal?: boolean;
}>;
/**
 * Creates the state store.
 *
 * @param initialState Initial state
 * @param options Parameters for the store
 */
declare function createStore<State>(initialState: State, options?: StoreOptions<State>): Store<State>;
/** Creates StateUpdates for updating the store by provided state mutations */
declare function createStoreUpdates<State, Updates extends StateUpdates<State>>(storeUpdate: Store<State>['update'], stateUpdates: Updates): StoreUpdates<State, Updates>;
/** Creates a proxy for the store with "updates" to change a state by provided mutations */
declare function withStoreUpdates<State, Updates extends StateUpdates<State> = StateUpdates<State>>(store: Store<State>, updates: Updates): StoreWithUpdates<State, Updates>;

/**
 * A controller-like boundary for effects and business logic.
 *
 * `Scope` collects all subscriptions which are made by child entities and provides
 * `destroy()` method to unsubscribe from them.
 */
type Scope = Controller<{
    /**
     * Register subscription-like or teardown function to be called with
     * `destroy()` method.
     */
    add: (teardown: TeardownLogic) => void;
    /**
     * Register a teardown function to be called with `destroy()` method.
     */
    onDestroy: (teardown: () => void) => void;
    /**
     * Creates a store which will be destroyed with the scope.
     *
     * @param initialState Initial state
     * @param options Parameters for the store
     */
    createStore<State>(initialState: State, options?: StoreOptions<State>): Store<State>;
    /**
     * Creates a controller which will be destroyed with the scope.
     */
    createController: <ControllerProps extends AnyObject = AnyObject>(factory: () => Controller<ControllerProps>) => Controller<ControllerProps>;
    /**
     * Creates an effect which will be destroyed with the scope.
     */
    createEffect: <Event = void, Result = void, ErrorType = Error>(handler: EffectHandler<Event, Result>, options?: EffectOptions<Event, Result>) => Effect<Event, Result, ErrorType>;
    /**
     * Creates an effect which handles `source` by `handler`, and it will be
     * destroyed with the scope.
     */
    handle: <Event, Result = void, ErrorType = Error>(source: Observable<Event> | Action<Event> | Query<Event>, handler: EffectHandler<Event, Result>, options?: EffectOptions<Event, Result>) => Effect<Event, Result, ErrorType>;
    /**
     * Subscribes to the `source` observable until the scope will be destroyed.
     */
    subscribe: {
        <T>(source: Observable<T>): Subscription;
        <T>(source: Observable<T>, next: (value: T) => void): Subscription;
        <T>(source: Observable<T>, observer: Partial<Observer<T>>): Subscription;
    };
}>;
/**
 * `ExternalScope` and `Scope` types allow to distinct which third-party code can invoke `destroy()` method.
 */
type ExternalScope = Omit<Scope, 'destroy'>;
/**
 * Creates `Scope` instance.
 */
declare function createScope(): Scope;

declare function createController<Service extends AnyObject>(factory: (scope: Scope) => Service & {
    destroy?: () => void;
}): Controller<Service>;
type ControllerFactory<Service extends AnyObject> = (container: Container) => Controller<Service>;
declare type DependencyProps = {
    [key: string]: unknown;
};
declare type TokenProps<Props extends DependencyProps> = {
    [K in keyof Props]: Token<Props[K]>;
};
declare function declareController<Dependencies extends DependencyProps, Service extends AnyObject>(tokens: TokenProps<Dependencies>, factory: (deps: Dependencies, scope: Scope) => Service): ControllerFactory<Service>;
type ViewControllerFactory<Service extends AnyObject, Params extends Query<unknown>[]> = (container: Container, ...params: Params) => Controller<Service>;
type InferredService<Factory> = Factory extends ViewControllerFactory<infer Service, infer Params> ? Service : never;
declare function declareViewController<Service extends AnyObject, Params extends Query<unknown>[]>(factory: (scope: Scope, ...params: Params) => Service): ViewControllerFactory<Service, Params>;
declare function declareViewController<Dependencies extends DependencyProps, Service extends AnyObject, Params extends Query<unknown>[]>(tokens: TokenProps<Dependencies>, factory: (deps: Dependencies, scope: Scope) => ((scope: Scope, ...params: Params) => Service) | Service): ViewControllerFactory<Service, Params>;

/**
 * Creates a new `Query` which maps a source value by the provided mapping
 * function.
 *
 * @param query source query
 * @param mapper value mapper
 */
declare function mapQuery<T, R>(query: Query<T>, mapper: (value: T) => R): Query<R>;
/**
 * Creates a new `Query` which takes the latest values from source queries
 * and merges them into a single value.
 *
 * @param queries source queries
 * @param merger value merger
 */
declare function mergeQueries<Values extends unknown[], Result>(queries: [
    ...{
        [K in keyof Values]: Query<Values[K]>;
    }
], merger: (...values: Values) => Result): Query<Result>;

/**
 * Creates a query for the specified BehaviourSubject
 *
 * @example
 * ```ts
 * const source = new BehaviourSubject(1);
 * const query = queryBehaviourSubject(source);

 * const result = compute((get) => get(query) + 1);

 * expect(result.get()).toBe(2)
 * ```
 */
declare function queryBehaviourSubject<T>(source: BehaviorSubject<T>): Query<T>;

/**
 * Creates a deferred or transformed view of the store.
 */
declare function pipeStore<T>(store: Store<T>, operator: MonoTypeOperatorFunction<T>): StoreQuery<T>;

type StoreEvent<State> = {
    type: 'created';
    store: Store<State>;
} | {
    type: 'destroyed';
    store: Store<State>;
} | {
    type: 'mutation';
    store: Store<State>;
    mutation: StateMutation<State>;
    nextState: State;
    prevState: State;
} | {
    type: 'updated';
    store: Store<State>;
    nextState: State;
    prevState: State;
};

type StateMutationMetadata = Readonly<{
    name?: string;
}>;
declare function getStateMutationMetadata(mutation: StateMutation<unknown>): StateMutationMetadata;

type StoreExtensionApi = Readonly<{
    getStateMutationMetadata: typeof getStateMutationMetadata;
}>;
type StoreExtension<State> = (api: StoreExtensionApi) => {
    onStoreEvent?: (event: StoreEvent<State>) => void;
};
declare function registerStoreExtension(extension: StoreExtension<unknown>): Subscription;

declare function createStoreLoggerExtension(logger: typeof console.log): StoreExtension<unknown>;

type StoreDeclaration<State, Updates extends StateUpdates<State> = StateUpdates<State>> = Readonly<{
    initialState: State;
    updates: Updates;
    options?: StoreOptions<State>;
}>;
type FactoryStateArg<State> = (State extends Function ? never : State) | StateMutation<State>;
type DeclaredStoreFactory<State, Updates extends StateUpdates<State>> = {
    (initialState?: FactoryStateArg<State>, options?: StoreOptions<State>): StoreWithUpdates<State, Updates>;
    updates: Updates;
};
/**
 * declare the base interface for create store
 * @example
```ts
type State = {
  id: string;
  name: string;
  isAdmin: boolean
};
const initialState: State = {
  id: '',
  name: '',
  isAdmin: false
};
const createUserStore = declareStore({
  initialState,
  updates: {
    setId: (id: string) => (state) => {
      return {
        ...state,
        id: id,
      };
    },
    setName: (name: string) => (state) => {
      return {
        ...state,
        name: name,
      };
    },
    update: (id: string name: string) => (state) => {
      return {
        ...state,
        id: id,
        name: name,
      };
    },
    setIsAdmin: () => (state) => {
      return {
        ...state,
        isAdmin: true,
      };
    },
  },
});

const userStore1 = createUserStore({ id: '1', name: 'User 1', isAdmin: false });

const userStore2 = createUserStore({ id: '2', name: 'User 2', isAdmin: true });

// OR

const users = [
  createUserStore({id: 1, name: 'User 1'}),
  createUserStore({id: 2, name: 'User 2'}),
]

userStore1.updates.setName('User from store 1');

assets.isEqual(userStore1.get().name, 'User from store 1')

assets.isEqual(userStore2.get().name, 'User 2')

// type of createUserStore
type UserStore = ReturnType<typeof createUserStore>;
```
 */
declare function declareStore<State, Updates extends StateUpdates<State> = StateUpdates<State>>(declaration: StoreDeclaration<State, Updates>): DeclaredStoreFactory<State, Updates>;

type Comparator<T> = (a: T, b: T) => boolean;
/**
 * This function returns a current value of a provided Query and registers it as a dependency for computation.
 */
type ComputationResolver = {
    <T>(query: Query<T>): T;
    <T, R>(query: Query<T>, selector: (value: T) => R): R;
};
/**
 * This function calculates a value from external sources or resolved sources by a provided "resolver".
 */
type Computation<T> = (resolver: ComputationResolver) => T;
/**
 * Options for "compute()" function
 */
type ComputationOptions<T> = {
    /** A custom comparator to differ complex values */
    comparator?: Comparator<T>;
    /** Explicitly dependencies for refreshing calculations */
    dependencies?: Query<unknown>[];
};
/**
 * Creates a computable query which calculates its values by provided "computation" function and dependencies.
 *
 * Rules of "Computation" function:
 * - it must have no side effects
 * - it recalculates only when specified dependencies are updated
 * - its "formula" may have other sources of values, however they don't trigger updates
 *
 * "Computation" function provides a resolver for using a dependency withing a calculation expression.
 *
 * Dependency can be declared explicitly as an array by the second argument. It has the following advantages:
 * - Faster dependency subscription
 * - Ability to specify extra queries or observables as dependencies
 *
 * A custom value comparator can be specified by "options" object as the second argument.
 * It helps to decide if a new value differs from a previous one in complex cases.
 *
 * @example
 * ```ts
 * const greeting = createStore('Hello');
 * const username = createStore('World');
 *
 * // Dependency are implicitly resolved
 * const message = compute((get) => get(greeting) + ' ' + get(username) + '!');
 *
 * // Dependency declared explicitly
 * const messageUppercase = compute(() => message.get().toUpperCase(), [message]);
 *
 * expect(message.get()).toBe('Hello World!');
 * expect(messageUppercase.get()).toBe('HELLO WORLD!');
 * ```
 */
declare const compute: {
    <T>(computation: Computation<T>, dependencies?: Query<unknown>[]): Query<T>;
    <T>(computation: Computation<T>, options?: ComputationOptions<T>): Query<T>;
};

export { Action, Computation, ComputationOptions, ComputationResolver, Controller, ControllerFactory, DeclaredStoreFactory, Effect, EffectController, EffectError, EffectErrorOrigin, EffectEventProject, EffectHandler, EffectNotification, EffectOptions, EffectPipeline, EffectResult, EffectState, ExternalScope, GLOBAL_EFFECT_UNHANDLED_ERROR$, InferredService, InternalStoreOptions, OBJECT_COMPARATOR, Query, Scope, StateMutation, StateMutationMetadata, StateUpdates, Store, StoreDeclaration, StoreEvent, StoreExtension, StoreOptions, StoreQuery, StoreUpdate, StoreUpdateFunction, StoreUpdates, StoreWithUpdates, ViewControllerFactory, compute, createAction, createController, createEffect, createEffectController, createScope, createStore, createStoreLoggerExtension, createStoreUpdates, declareController, declareStateUpdates, declareStore, declareViewController, mapQuery, mergeQueries, pipeStateMutations, pipeStore, queryBehaviourSubject, registerStoreExtension, withStoreUpdates };
